#-------------------------------------------------------------.
# ベクトル/行列計算の便利機能.
#-------------------------------------------------------------.
import numpy as np
import math

#---------------------------------------.
# ゼロチェック.
# @param[in] v  実数値.
#---------------------------------------.
def isZero (v):
	minV = 1e-5
	if v > -minV and v < minV:
		return True
	return False

#---------------------------------------.
# vec3型でのゼロチェック.
# @param[in] v  (x, y, z)の3要素.
#---------------------------------------.
def isZero_vec3 (v):
	if isZero(v[0]) and isZero(v[1]) and isZero(v[2]):
		return True
	return False

#---------------------------------------.
# 単位行列を求める.
#---------------------------------------.
def identity_matrix4x4 ():
	return np.matrix(np.identity(4))

#---------------------------------------.
# 逆行列を計算.
# @param[in]   4x4行列.
#---------------------------------------.
def inverse_matrix4x4 (m):
	matrix_4x4 = np.matrix(m)
	return np.linalg.inv(matrix_4x4)

#---------------------------------------.
# vec3型を取得.
#---------------------------------------.
def identity_vec3 ():
	return np.array([0.0, 0.0, 0.0])

#---------------------------------------.
# vec3型を指定.
#---------------------------------------.
def set_vec3 (v):
	return np.array([v[0], v[1], v[2]])

#---------------------------------------.
# quaternion型を取得.
#---------------------------------------.
def identity_quaternion ():
	return np.array([1.0, 0.0, 0.0, 0.0])

#---------------------------------------.
# quaternion型を指定.
#---------------------------------------.
def set_quaternion (q):
	return np.array([q[0], q[1], q[2], q[3]])

#---------------------------------------.
# ベクトルの加算.
# @param[in] v1  (x, y, z)の3要素.
# @param[in] v2  (x, y, z)の3要素.
#---------------------------------------.
def add_vec3_vec3 (v1, v2):
	vec3_1 = np.array(v1)
	vec3_2 = np.array(v2)
	
	return vec3_1 + vec3_2

#---------------------------------------.
# ベクトルの減算.
# @param[in] v1  (x, y, z)の3要素.
# @param[in] v2  (x, y, z)の3要素.
#---------------------------------------.
def sub_vec3_vec3 (v1, v2):
	vec3_1 = np.array(v1)
	vec3_2 = np.array(v2)
	
	return vec3_1 - vec3_2

#---------------------------------------.
# 行列の指定.
# @param[in] matrix  4x4行列.
#---------------------------------------.
def set_matrix4x4 (matrix):
	return np.matrix(matrix)

#---------------------------------------.
# 行列とベクトルの乗算.
# @param[in] v  (x, y, z)の3要素.
#---------------------------------------.
def multiply_vec_matrix4x4 (v, matrix):
	vec = np.array(v)
	if len(v) == 3:
		vec = np.array([v[0], v[1], v[2], 1.0])
	matrix4x4 = np.matrix(matrix)
	ret = np.dot(vec, matrix4x4)

	return [ret[0,0], ret[0,1], ret[0,2], ret[0,3]]

#---------------------------------------.
# 行列同士の乗算.
#---------------------------------------.
def multiply_matrix4x4_matrix4x4 (matrix1, matrix2):
	matrix4x4_1 = np.array(matrix1)
	matrix4x4_2 = np.array(matrix2)
	return np.dot(matrix4x4_1, matrix4x4_2)

#---------------------------------------.
# ベクトルの長さを計算.
#---------------------------------------.
def length_vec3 (v):
	vec3 = np.array(v)
	return np.linalg.norm(vec3)

#---------------------------------------.
# 単位ベクトルを計算.
#---------------------------------------.
def normalize_vec3 (v):
	vec3 = np.array(v)
	len = np.linalg.norm(vec3)
	if isZero(len):
		return [0, 0, 0]
	return vec3 / len

def normalize_vec4 (v):
	vec4 = np.array(v)
	len = np.linalg.norm(vec4)
	if isZero(len):
		return [0, 0, 0, 0]
	return vec4 / len

#---------------------------------------.
# 内積を計算.
#---------------------------------------.
def dot_vec3 (v1, v2):
	vec3_1 = np.array(v1)
	vec3_2 = np.array(v2)
	return np.dot(vec3_1, vec3_2)

#---------------------------------------.
# 外積を計算.
#---------------------------------------.
def cross_vec3 (v1, v2):
	vec3_1 = np.array(v1)
	vec3_2 = np.array(v2)
	return np.cross(vec3_1, vec3_2)

#---------------------------------------.
# 移動行列を求める.
#---------------------------------------.
def translate_matrix4x4 (trans):
	m = identity_matrix4x4()
	m[3,0] = trans[0]
	m[3,1] = trans[1]
	m[3,2] = trans[2]
	return m

#---------------------------------------.
# 拡大縮小行列を求める.
#---------------------------------------.
def scale_matrix4x4 (scale):
	m = identity_matrix4x4()
	m[0,0] = scale[0]
	m[1,1] = scale[1]
	m[2,2] = scale[2]
	return m

#---------------------------------------.
# 回転行列を求める.
#---------------------------------------.
def rotate_matrix4x4 (rotate):
	m = identity_matrix4x4()

	s = math.sin(rotate[0])
	c = math.cos(rotate[0])

	rx = identity_matrix4x4()
	rx[1,1] = c
	rx[1,2] = s
	rx[2,1] = -s
	rx[2,2] = c

	s = math.sin(rotate[1])
	c = math.cos(rotate[1])
	ry = identity_matrix4x4()
	ry[0,0] = c
	ry[0,2] = -s
	ry[2,0] = s
	ry[2,2] = c

	s = math.sin(rotate[2])
	c = math.cos(rotate[2])
	rz = identity_matrix4x4()
	rz[0,0] = c
	rz[0,1] = s
	rz[1,0] = -s
	rz[1,1] = c

	m = np.dot(rx, ry)
	m = np.dot(m, rz)
	return m

#---------------------------------------.
# ベクトルaをbに向かせる回転行列を求める.
#---------------------------------------.
def rotate_a_to_b_matrix4x4 (a, b):
	m = identity_matrix4x4()
	a = np.array([_a[0], _a[1], _a[2]])
	b = np.array([_b[0], _b[1], _b[2]])
	return quaternion_matrix4x4( quaternion_a_b(a, b) )

#---------------------------------------.
# ベクトルaをbに向かせるquaternionを求める.
#---------------------------------------.
def quaternion_a_b (a, b):
	a = normalize_vec3(a)
	b = normalize_vec3(b)
	q = [0.0, 0.0, 0.0, 0.0]

	c = cross_vec3(b, a)
	d = -length_vec3(c)
	c = normalize_vec3(c)

	epsilon = 0.0002
	ip = dot_vec3(a, b)
	if -epsilon < d or 1.0 < ip:
		if ip < (epsilon - 1.0):
			a2 = [-a[1], a[2], a[0]]
			c = normalize_vec3(cross_vec3(a2, a))
			q[0] = 0.0
			q[1] = c[0]
			q[2] = c[1]
			q[3] = c[2]
		else:
			q = np.array([1.0, 0.0, 0.0, 0.0])
	else:
		e = c * math.sqrt(0.5 * (1.0 - ip))
		q[0] = math.sqrt(0.5 * (1.0 + ip))
		q[1] = e[0]
		q[2] = e[1]
		q[3] = e[2]
	return q

#---------------------------------------.
# quaternionの逆数を計算.
#---------------------------------------.
def inverse_quaternion (q):
	m = quaternion_matrix4x4(q)
	m = inverse_matrix4x4(m)
	return matrix4x4_quaternion(m)

#---------------------------------------.
# quaternionから4x4行列を求める.
# @param[in]  quaternion値。(w,x,y,z)として入っている.
# @return  4x4の回転行列.
#---------------------------------------.
def quaternion_matrix4x4 (q):
	m = identity_matrix4x4()

	# q(w,x,y,z)から(x,y,z,w)に並べ替え.
	q2 = range(4)
	q2[0] = q[1]
	q2[1] = q[2]
	q2[2] = q[3]
	q2[3] = q[0]

	m[0,0] = q2[3]*q2[3] + q2[0]*q2[0] - q2[1]*q2[1] - q2[2]*q2[2]
	m[0,1] = 2.0 * q2[0] * q2[1] - 2.0 * q2[3] * q2[2]
	m[0,2] = 2.0 * q2[0] * q2[2] + 2.0 * q2[3] * q2[1]
	m[0,3] = 0.0

	m[1,0] = 2.0 * q2[0] * q2[1] + 2.0 * q2[3] * q2[2]
	m[1,1] = q2[3] * q2[3] - q2[0] * q2[0] + q2[1] * q2[1] - q2[2] * q2[2]
	m[1,2] = 2.0 * q2[1] * q2[2] - 2.0 * q2[3] * q2[0]
	m[1,3] = 0.0

	m[2,0] = 2.0 * q2[0] * q2[2] - 2.0 * q2[3] * q2[1]
	m[2,1] = 2.0 * q2[1] * q2[2] + 2.0 * q2[3] * q2[0]
	m[2,2] = q2[3] * q2[3] - q2[0] * q2[0] - q2[1] * q2[1] + q2[2] * q2[2]
	m[2,3] = 0.0

	m[3,0] = 0.0
	m[3,1] = 0.0
	m[3,2] = 0.0
	m[3,3] = q2[3] * q2[3] + q2[0] * q2[0] + q2[1] * q2[1] + q2[2] * q2[2]

	k = m[3,3]
	for i in range(3):
		for j in range(3):
			m[i,j] /= k

	m[3,3] = 1.0

	return m

#---------------------------------------.
# 4x4行列をスケール、せん断、回転、移動要素に分解.
# scale, shear, rotate, translateは、
# あらかじめ np.array([0.0, 0.0, 0.0]) として初期化しておくこと.
# @param[in]   matrix  4x4行列.
# @param[out]  scale      スケール値が返る。(x, y, z)要素.
# @param[out]  shear      せん断値が返る。(x, y, z)要素.
# @param[out]  rotate     回転値が返る。(x, y, z)要素.
# @param[out]  translate  移動値が返る。(x, y, z)要素.
#---------------------------------------.
def unmatrix_matrix4x4 (matrix, scale, shear, rotate, translate):
	#scale     = np.array([0.0, 0.0, 0.0])
	#shear     = np.array([0.0, 0.0, 0.0])
	#rotate    = np.array([0.0, 0.0, 0.0])
	#translate = np.array([0.0, 0.0, 0.0])

	m      = np.matrix(matrix)
	locmat = m
	row    = np.matrix(np.identity(3))
	p      = np.array([0.0, 0.0, 0.0])
	a      = 0.0

	# 正規化.
	for i in range(4):
		for j in range(4):
			locmat[i,j] /= locmat[3,3]

	pmat = locmat

	for i in range(3):
		pmat[i,3] = 0.0
	
	pmat[3,3] = 1.0;

	for i in range(3):
		translate[i] = locmat[3,i]
		locmat[3,i]  = 0.0

	for i in range(3):
		row[i,0] = locmat[i,0]
		row[i,1] = locmat[i,1]
		row[i,2] = locmat[i,2]

	v = [ row[0,0], row[0,1], row[0,2] ]
	scale[0] = length_vec3(v)
	v = normalize_vec3(v)
	row[0,0] = v[0]
	row[0,1] = v[1]
	row[0,2] = v[2]

	v1 = [ row[0,0], row[0,1], row[0,2] ]
	v2 = [ row[1,0], row[1,1], row[1,2] ]
	shear[2] = dot_vec3(v1, v2)

	row[1,0] -= row[0,0] * shear[2]
	row[1,1] -= row[0,1] * shear[2]
	row[1,2] -= row[0,2] * shear[2]

	v = [ row[1,0], row[1,1], row[1,2] ]
	scale[1] = length_vec3(v)
	v = normalize_vec3(v)
	row[1,0] = v[0]
	row[1,1] = v[1]
	row[1,2] = v[2]

	shear[2] = shear[2] / scale[1]

	v1 = [ row[0,0], row[0,1], row[0,2] ]
	v2 = [ row[1,0], row[1,1], row[1,2] ]
	shear[1] = dot_vec3(v1, v2)

	row[2,0] -= row[0,0] * shear[1]
	row[2,1] -= row[0,1] * shear[1]
	row[2,2] -= row[0,2] * shear[1]

	v1 = [ row[1,0], row[1,1], row[1,2] ]
	v2 = [ row[2,0], row[2,1], row[2,2] ]
	shear[0] = dot_vec3(v1, v2)

	row[2,0] -= row[1,0] * shear[0]
	row[2,1] -= row[1,1] * shear[0]
	row[2,2] -= row[1,2] * shear[0]

	v = [ row[2,0], row[2,1], row[2,2] ]
	scale[2] = length_vec3(v)
	v = normalize_vec3(v)
	row[2,0] = v[0]
	row[2,1] = v[1]
	row[2,2] = v[2]

	shear[1] = shear[1] / scale[2]
	shear[0] = shear[0] / scale[2]

	v1 = [ row[1,0], row[1,1], row[1,2] ]
	v2 = [ row[2,0], row[2,1], row[2,2] ]
	dd = cross_vec3(v1, v2)
	p[0] = dd[0]
	p[1] = dd[1]
	p[2] = dd[2]

	v1 = [ row[0,0], row[0,1], row[0,2] ]
	v2 = [ p[0], p[1], p[2] ]
	fVal = dot_vec3(v1, v2)

	if fVal < 0.0:
		for i in range(3):
			scale[i] *= -1.0
			row[i,0] *= -1.0
			row[i,1] *= -1.0
			row[i,2] *= -1.0

	rotate[1] = math.asin(-row[0,2])
	a = math.cos(rotate[1])

	epsilon = 0.0002
	if a < epsilon or epsilon < a:
		rotate[0] = math.atan2(row[1,2], row[2,2])
		rotate[2] = math.atan2(row[0,1], row[0,0])
	else:
		if 0.0 < rotate[1]:
			rotate[0] = math.atan2(row[1,0], row[1,1])
		else:
			rotate[0] = -math.atan2(row[1,0], row[1,1])
		rotate[2] = 0.0

#---------------------------------------.
# quaternionの計算.
# @param[in] w    w値.
# @param[in] v    (x, y, z)値.
# @return   (w,x,y,z)としてquaternion値が返る.
#---------------------------------------.
def quaternion_w_xyz (w, v):
	q = np.array([w, v[0], v[1], v[2]])
	return q

#---------------------------------------.
# quaternionの乗算.
# @param[in] q1  quaternion 1
# @param[in] q2  quaternion 2
#---------------------------------------.
def multiply_quaternion_quaternion (q1, q2):
	u = np.array([q1[1], q1[2], q1[3]])
	v = np.array([q2[1], q2[2], q2[3]])
	q = quaternion_w_xyz(q1[0] * q2[0] - dot_vec3(u, v), v * q1[0] + u * q2[0] + cross_vec3(u, v))
	return normalize_vec4(q)

#---------------------------------------.
# 回転行列からquaternionを求める.
# @param[in] matrix   4x4回転行列.
# @return   (w,x,y,z)としてquaternion値が返る.
#---------------------------------------.
def matrix4x4_quaternion (matrix):
	scale     = identity_vec3()
	shear     = identity_vec3()
	rotate    = identity_vec3()
	translate = identity_vec3()
	unmatrix_matrix4x4(matrix, scale, shear, rotate, translate)

	q = np.array([0.0, 0.0, 0.0, 0.0])

	q1 = quaternion_w_xyz(math.cos(rotate[0] / 2.0), [math.sin(-rotate[0] / 2.0), 0.0, 0.0])
	q2 = quaternion_w_xyz(math.cos(rotate[1] / 2.0), [0.0, math.sin(-rotate[1] / 2.0), 0.0])
	q3 = quaternion_w_xyz(math.cos(rotate[2] / 2.0), [0.0, 0.0, math.sin(-rotate[2] / 2.0)])

	q = multiply_quaternion_quaternion(q1, q2)
	q = multiply_quaternion_quaternion(q, q3)
	return q

#---------------------------------------.
# ベジェ上の位置を計算.
# @param[in] v1Pos  ベジェ開始点.
# @param[in] v1Out  ベジェ開始点のout handle.
# @param[in] v2In   ベジェ終了点のin handle.
# @param[in] v2Pos  ベジェ終了点.
# @param[in] fPos   ベジェ上の位置 (0.0 - 1.0).
# @return  ベジェ上の位置.
#---------------------------------------.
def get_bezier_point (v1Pos, v1Out, v2In, v2Pos, fPos):
	fMin = 1e-6

	rPos = [0.0, 0.0, 0.0]
	cPos = []
	cPos.append([v1Pos[0], v1Pos[1], v1Pos[2]])
	cPos.append([v1Out[0], v1Out[1], v1Out[2]])
	cPos.append([v2In[0],  v2In[1],  v2In[2]])
	cPos.append([v2Pos[0], v2Pos[1], v2Pos[2]])

	fPos2 = float(fPos)
	fPos2 = max(0.0, fPos2)
	fPos2 = min(1.0, fPos2)
	
	if isZero(fPos2):
		rPos = cPos[0]
		return rPos

	if isZero(fPos2 - 1.0):
		rPos = cPos[3]
		return rPos

	# ベジェ曲線の計算.
	t   = fPos2
	t2  = 1.0 - t
	t2d = t2 * t2
	td  = t * t
	b1  = t2d * t2
	b2  = 3.0 * t * t2d
	b3  = 3.0 * td * t2
	b4  = t * td

	for i in range(3):
		rPos[i] = b1 * cPos[0][i] + b2 * cPos[1][i] + b3 * cPos[2][i] + b4 * cPos[3][i]

	return rPos
